package edu.stanford.nlp.fsm;

import edu.stanford.nlp.util.Scored;

import java.util.*;

/**
 * DFSAState
 * <p/>
 * Class for representing the state of a deterministic finite state
 * automaton without epsilon transitions.
 *
 * @author Dan Klein
 * @version 12/14/2000
 */
public final class DFSAState<T,S> implements Scored {
  public S stateID;
  public Map<T,DFSATransition<T,S>> inputToTransition;
  public boolean accepting;
  public DFSA<T,S> dfsa;

  public double score;

  public double score() {
    return score;
  }

  public void setScore(double score) {
    this.score = score;
  }


  public DFSA dfsa() {
    return dfsa;
  }

  public void setStateID(S stateID) {
    this.stateID = stateID;
  }

  public S stateID() {
    return stateID;
  }

  public void addTransition(DFSATransition<T,S> transition) {
    inputToTransition.put(transition.input(), transition);
  }

  public DFSATransition<T,S> transition(T input) {
    return (DFSATransition<T,S>) inputToTransition.get(input);
  }

  public Collection<DFSATransition<T,S>> transitions() {
    return inputToTransition.values();
  }

  public Set<T> continuingInputs() {
    return inputToTransition.keySet();
  }

  public Set<DFSAState<T,S>> successorStates() {
    Set<DFSAState<T,S>> successors = new HashSet<DFSAState<T,S>>();
    Collection transitions = inputToTransition.values();
    for (Iterator transitionI = transitions.iterator(); transitionI.hasNext();) {
      DFSATransition<T,S> transition = (DFSATransition<T,S>) transitionI.next();
      successors.add(transition.getTarget());
    }
    return successors;
  }

  public void setAccepting(boolean accepting) {
    this.accepting = accepting;
  }

  public boolean isAccepting() {
    return accepting;
  }

  public boolean isContinuable() {
    return !inputToTransition.isEmpty();
  }

  public String toString() {
    return stateID.toString();
  }

  private int hashCodeCache = 0;

  public int hashCode() {
    if (hashCodeCache == 0) {
      hashCodeCache = stateID.hashCode() ^ dfsa.hashCode();
    }
    return hashCodeCache;
  }

  public boolean equals(Object o) {
    if (!(o instanceof DFSAState)) {
      return false;
    }
    DFSAState s = (DFSAState) o;
    if (dfsa.equals(s.dfsa) && stateID.equals(s.stateID)) {
      //accepting == s.accepting &&
      //inputToTransition.equals(s.inputToTransition))
      return true;
    }
    return false;
  }

  public Set statesReachable() {
    Set visited = new HashSet();
    List toVisit = new ArrayList();
    toVisit.add(this);
    exploreStates(toVisit, visited);
    return visited;
  }

  private void exploreStates(List toVisit, Set visited) {
    while (!toVisit.isEmpty()) {
      DFSAState state = (DFSAState) toVisit.get(toVisit.size() - 1);
      toVisit.remove(toVisit.size() - 1);
      if (!visited.contains(state)) {
        toVisit.addAll(state.successorStates());
        visited.add(state);
      }
    }
  }

  public DFSAState(S id, DFSA<T,S> dfsa) {
    this.dfsa = dfsa;
    this.stateID = id;
    this.accepting = false;
    this.inputToTransition = new HashMap<T,DFSATransition<T,S>>();
    this.score = Double.NEGATIVE_INFINITY;
  }

  public DFSAState(S id, DFSA<T,S> dfsa, double score) {
    this(id,dfsa);
    setScore(score);
  }



}
